#include "SubmatClass.h"
#include "misc.h"



SubmatClass_t::SubmatClass_t(int K) {
  /* define alphabet and create index map*/
  alphabet ="ARNDCQEGHILKMFPSTWYV";
  for (int i = 0; i < alphabet.length(); i++) {
      aa2idx.insert (pair<char,int>(alphabet[i],i));
  }

  this->K  = 1;
  S        = readMatrix("data/S_double_precision.txt");
  D        = readMatrix("data/D_double_precision.txt");
  invS     = readMatrix("data/S_inv_double_precision.txt");
  Mpow1    = matrixMultiply(S,matrixMultiply(D,invS));
  exp(K);
  dirichletpriors = readDirichletPriors("data/DirichletPriorsGivenN.txt");
}

vector<vector<double> > SubmatClass_t::readMatrix(string fn) {
    ifstream infile(fn.c_str(),ios::in);
    assert(infile);
    vector<vector<double> > mat;
    size_t nRows = alphabet.size();
    size_t nCols = nRows;
    for (size_t i = 0; i < nRows; i++) {
        vector<double> row;
        double tmp;
        for (size_t j = 0; j < nCols; j++) {
            infile >> tmp;
            row.push_back(tmp);
        }
        mat.push_back(row);
    }
    infile.close();
    return mat;
}

vector<vector<DirichletClass_t> > SubmatClass_t::readDirichletPriors(string fn) {
    ifstream infile(fn.c_str(),ios::in);
    assert(infile);
    vector<DirichletClass_t> priors_Mstate;
    vector<DirichletClass_t> priors_Istate;
    char buff[1000];
    int cntr = 1;
    while (!infile.eof()) {
        infile.getline(buff,1000);
        if (infile.eof()) break;

        if (buff[0] == '#') continue; //ignore comment line
        int pamK;
        double m1,m2, i1,i2,i3;
        vector<double> Malphas(2);
        vector<double> Ialphas(3);
        stringstream ss;
        ss << buff;
        ss >> pamK;
        assert(pamK == cntr++);

        ss >> Malphas[0];
        ss >> Malphas[1];
        ss >> Ialphas[0];
        ss >> Ialphas[1];
        ss >> Ialphas[2];
        DirichletClass_t objM(2,Malphas);
        DirichletClass_t objI(3,Ialphas);
        priors_Mstate.push_back(objM);
        priors_Istate.push_back(objI);
    }
    infile.close();
    assert(priors_Mstate.size() == 1000);
    assert(priors_Istate.size() == 1000);

    vector<vector<DirichletClass_t> > priors;
    priors.push_back(priors_Mstate);
    priors.push_back(priors_Istate);
    assert(priors.size()==2);  
    return priors;
}

void SubmatClass_t::exp(int K) {
    assert(K>0);
    assert(K<=1000);
    this->K = K;
    if (K == 1)  {
        MpowK = Mpow1;
    }
    else {
        vector<vector<double> > DpowK = D;
        for (size_t i = 0;  i < D.size(); i++) {
            DpowK[i][i] = pow(D[i][i],K);
        }
        MpowK = matrixMultiply(S,matrixMultiply(DpowK,invS));
    }
  /* normalize MpowK matrix so that each col adds to 1 */
  normalizeMpowK();
  return;
}

double SubmatClass_t::pr(char x, char y) {
    map<char,int>::iterator it;
    it = aa2idx.find(x);
    assert(it != aa2idx.end());
    it = aa2idx.find(y);
    assert(it != aa2idx.end());
    return MpowK[aa2idx[x]][aa2idx[y]];
}

/* The substitution matrix stores conditional probabilities,
 * and is    NOT SYMMETRIC!!!.. This interface is returning
 * the prob of x given y.*/
double SubmatClass_t::pr(unsigned int x, unsigned int y) {
    static int n = alphabet.length();
    assert(x < n);
    assert(y < n);
    return MpowK[aa2idx[x]][aa2idx[y]];
}

string SubmatClass_t::getAlphabet() {
    return alphabet;
}

double SubmatClass_t::getS(int x,int y) {
    static int n = alphabet.length();
    assert(x < n && x >= 0);
    assert(y < n && y >= 0);
    return S[x][y];
}

double SubmatClass_t::getSinv(int x,int y) {
    static int n = alphabet.length();
    assert(x < n && x >= 0);
    assert(y < n && y >= 0);
    return invS[x][y];
}

double SubmatClass_t::getD(int x,int y) {
    static int n = alphabet.length();
    assert(x < n && x >= 0);
    assert(y < n && y >= 0);
    return D[x][y];
}

int SubmatClass_t::getIndx(char x) {
    assert(alphabet.find(x)!=string::npos);
    return aa2idx[x];
}

int SubmatClass_t::getK() {
    return K;
}

void SubmatClass_t::normalizeMpowK() {
    assert(K>0);

    int n = alphabet.length();

    /* When K is small (eg K=1) elements of MpowK approach 0 from the -ve side
     * due to numerical issues that result from floating point matrix multiplication.
     * So, first convert values <0 to 0 */
    for (int i = 0;  i < n; i++) {
        for (int j = 0;  j < n; j++) {
            if (MpowK[i][j] < 0) MpowK[i][j] = 0.0;
        }
    }

    //find col sums
    vector<double> colsum;
    for (int j = 0;  j < n; j++) {
        double tmpsum = 0;
        for (int i = 0;  i < n; i++) {
            tmpsum += MpowK[i][j];
        }
        colsum.push_back(tmpsum);
    }

    //matrix rows
    for (int i = 0;  i < n; i++) {
        for (int j = 0;  j < n; j++) {
            MpowK[i][j] /= colsum[j];
        }
    }
}

DirichletClass_t SubmatClass_t::getDirichlet(string state) {
    //only allowed arguments are "match" and "insert"
    if (state.compare("match") == 0) return dirichletpriors[0][K-1];
    else if (state.compare("insert") == 0) return dirichletpriors[1][K-1];
    else assert(!"getDirichlet state string invalid") ;
}

void SubmatClass_t::printMpowK() {
    assert(K>0);
    cout << "Printing details of M^K when K=" << K << endl;

    int n = alphabet.length();
    //find row sums
    vector<double> rowsum;
    for (int i = 0;  i < n; i++) {
        double tmpsum = 0;
        for (int j = 0;  j < n; j++) {
            tmpsum += MpowK[i][j];
        }
        rowsum.push_back(tmpsum);
    }
    //find col sums
    vector<double> colsum;
    for (int j = 0;  j < n; j++) {
        double tmpsum = 0;
        for (int i = 0;  i < n; i++) {
            tmpsum += MpowK[i][j];
        }
        colsum.push_back(tmpsum);
    }

    int w = 7, p = 3;
    cout << fixed;
    //alphabet row
    cout << setw(w) << " ";
    for (int i = 0;  i < n; i++) {
        cout << setw(w) << alphabet[i];
    }
    cout << " | ";
    cout << setw(w) << "rowsum";
    cout << endl;

    //matrix rows
    for (int i = 0;  i < n; i++) {
        cout << setw(w) << alphabet[i];
        for (int j = 0;  j < n; j++) {
            cout << setw(w) << setprecision(p) << MpowK[i][j];
        }
        cout << " | ";
        cout << setw(w) << setprecision(p) << rowsum[i];
        cout << endl;
    }
    //col sum line
    cout << "------";
    cout << endl;
    cout << setw(w) << "colsum";
    for (int j = 0;  j < n; j++) {
        cout << setw(w) << setprecision(p) << colsum[j];
    }
    cout << endl;
    cout << endl;

    cout << "Printing associated Dirichlet priors for K=" << K << endl;
    cout << "Params for M-state priors: " << endl;
    dirichletpriors[0][K-1].printParams();
    cout << "Params for I-state priors: " << endl;
    dirichletpriors[1][K-1].printParams();
}
